/* Analyse von PARMOD-Rechnungen */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

static char Path[256];
static char Mode[40], Label[80];
static FILE *Log;
static char Buf[4000];

static int HomA( int m )
  {
  int n, i, ii, j, k, l;
  float a, da, a1sum[100], a2sum[100];
  float sa, ma;
  char *pc;
  if (m > 100)  m = 100;
  l = strlen(Label);
  for (j=0; j<100; j++) { a1sum[j] = 0;  a2sum[j] = 0; }
  ii = 0;
  i = 0;
  n = 0;
  while (fgets(Buf, 4000, Log)) {
    n++;
    if (n < 4)  continue;
    if (strlen(Buf)<5+l || (Buf[l+4]!=':' && Buf[l+3]!=':')) {
      if (ii == 0)  continue;
      else  break;
      }
    if (l>0 && strncmp(Buf, Label, l))  continue;
    k = sscanf(Buf+l, "%d", &i);
    if (k != 1)  break;
    if (i != ii+1) {
      printf("sequence error (%d) in line %d\n", i, n);
      return -1;
      }
    ii = i;
    pc = strtok(Buf+5+l, " \t\r\n");
    for (j=0; j<m; j++) {
      if (!pc) {
        printf("value #%d in line %d not found!\n", j+1, n);
        return -3;
        }
      k = sscanf(pc, "%f", &a);
      if (k != 1)  break;
      a1sum[j] += a;
      a2sum[j] += a*a;
      pc = strtok(NULL, " \t\r\n");
      }
    }
  if (i < 2) {
    printf("only %d values!\n", i);
    return -2;
    }
  printf("n=%d\n", i);
  for (j=0; j<m; j++) {
    ma = a1sum[j]/i;
    sa = sqrt((a2sum[j] - a1sum[j]*a1sum[j]/i)/(i-1));
    da = sa/sqrt(i-1);
    printf("%4d  %7.1f  %5.3f  %7.1f  %7.1f\n",
      j+1, ma, sa/ma, ma-2*da, ma+2*da);
    }
  return 0;
  }

static int Histo( int m, float dc )
  {
  int n, i, ii, j, k, l, kmax;
  float a, da, a1sum[101];
  float sa, ma, s1, s2;
  char *pc;
  if (m > 100)  m = 100;
  l = strlen(Label);
  for (j=0; j<101; j++)  a1sum[j] = 0;
  ii = 0;
  i = 0;
  n = 0;
  s1 = 0;
  s2 = 0;
  kmax = 19;
  while (fgets(Buf, 4000, Log)) {
    n++;
    if (n < 4)  continue;
    if (strlen(Buf)<5+l || (Buf[l+4]!=':' && Buf[l+3]!=':')) {
      if (ii == 0)  continue;
      else  break;
      }
    if (l>0 && strncmp(Buf, Label, l))  continue;
    k = sscanf(Buf+l, "%d", &i);
    if (k != 1)  break;
    if (i != ii+1) {
      printf("sequence error (%d) in line %d\n", i, n);
      return -1;
      }
    ii = i;
    pc = strtok(Buf+5+l, " \t\r\n");
    for (j=1; j<=m; j++) {
      if (!pc) {
        printf("value #%d in line %d not found!\n", j+1, n);
        return -3;
        }
      if (j == m) {
        k = sscanf(pc, "%f", &a);
        if (k != 1)  break;
        s1 += a;
        s2 += a*a;
        k = a/dc;
        if (k > 100)  k = 100;
        if (k > kmax)  kmax = k;
        a1sum[k] += 1;
        }
      pc = strtok(NULL, " \t\r\n");
      }
    }
  if (i < 2) {
    printf("only %d values!\n", i);
    return -2;
    }
  ma = s1/i;
  sa = sqrt((s2 - i*ma*ma)/(i-1));
  da = sa/sqrt(i-1);
  printf("n=%d,  ma=%7.1f,  sa=%7.2f\n", i, ma, sa);
  for (k=0; k<=kmax; k++) {
    printf("%4d  %5.0f  %5.0f  %3.0f\n",
      k+1, k*dc, (k+1)*dc, a1sum[k]);
    }
  return 0;
  }

static int Ipol(float *xx, int nx, float **pyy)
  {
  int j, j1, j2, ny, n;
  float *yy, a, b, c, y, ym;
  ny = 2*nx + 1;
  yy = malloc(ny*sizeof(float));
  if (!yy)  return -1;
  for (j=0; j<ny; j++)  yy[j] = 0;
  for (j=1; j<ny; j+=2) yy[j] = 2*xx[j/2];
  for (n=0; n<10 ; n++) {
    j1 = 0;
    j2 = ny-1;
    for (j=j1; j<=j2; j+=2) {
      y = yy[j];
      if      (j == j1) { a = yy[j+1];  b = a; }
      else if (j == j2) { a = yy[j-1];  b = a; }
      else              { a = yy[j-1];  b = yy[j+1]; }
      ym = 0.5*(a + b);
      c = (a > b) ? b : a;
      c *= 0.5;
      if (c > 0.5*(ym-y))  c = 0.5*(ym-y);
      if (c < 0)  continue;
      yy[j] += c;
      if (j > j1)  yy[j-1] -= 0.5*c;
      if (j < j2)  yy[j+1] -= 0.5*c;
      }
    }
  *pyy = yy;
  return ny;
  }

static int SymVar(float *ww, int nw, float *pv)
  {
  int i;
  float sumw, sumv, y1, y2, dy, y1q, y2q, w1, w2, dv;
  sumw = 0;
  sumv = 0;
  dy = 0.5;
  for (i=1; i<nw; i++) {
    y1 = 0.5*(i - (nw-1)/2 - 1);
    y2 = y1 + dy;
    y1q = y1*y1;
    y2q = y2*y2;
    w1 = ww[i-1];
    w2 = ww[i];
    sumw += dy*0.5*(w1 + w2);
    dv = ((y2*w1-y1*w2)*(y2q*y2-y1q*y1)/3 + (w2-w1)*(y2q*y2q-y1q*y1q)/4)/dy;
    sumv += dv;
    }
  if (sumw > 0)  sumv /= sumw;
  *pv = sumv;
  return 0;
  }


static int Taylor( int m, int nt )
  {
  int n, i, ii, j, k, l, na, nb, it;
  float a, da, a0sum[100], a1sum[100], a2sum[100], aa[201], *pb;
  float sa, ma, a1, a2, q;
  char *pc;

  if (m > 100)  m = 100;
  if (nt > 99)  nt = 99;
  l = strlen(Label);

  for (it=1; it<=nt; it++) {
  a0sum[it] = 0;
  a1sum[it] = 0;
  a2sum[it] = 0;
  fseek(Log, 0, SEEK_SET);

  for (j=0; j<200; j++)  aa[j] = 0;
  ii = 0;
  i = 0;
  n = 0;
  while (fgets(Buf, 4000, Log)) {
    n++;
    if (n < 4)  continue;
    if (strlen(Buf)<5+l || Buf[l+4]!=':') {
      //if (ii == 0)  continue;
      //else  break;
      continue;
      }
    if (l>0 && strncmp(Buf, Label, l))  continue;
    k = sscanf(Buf+l, "%d", &i);
    if (k != 1)  break;
    j = ii*100 + it;
    if (i < j)  continue;
    if (i > j) {
      printf("sequence error (%d,%d) in line %d\n", i, j, n);
      return -1;
      }
    ii++;
    pc = strtok(Buf+5+l, " \t\r\n");
    for (j=0; j<m; j++) {
      if (!pc) {
        printf("value #%d in line %d not found!\n", j+1, n);
        return -3;
        }
      k = sscanf(pc, "%f", &a);
      if (k != 1)  break;
      aa[j] = a;
      pc = strtok(NULL, " \t\r\n");
      }
    nb = Ipol(aa, m, &pb);
    if (nb < 1) {
      printf("n=%d  it=%d  ii=%d  nb=%d\n", n, it, ii, nb);
      exit(1);
      }
    k = SymVar(pb, nb, &sa);
/*
printf("i=%4d  nb=%d  sa=%12.5e\n", i, nb, sa);
*/
    a0sum[it] += 1;
    a1sum[it] += sa;
    a2sum[it] += sa*sa;
    free(pb);
    }
  if (ii < 2) {
    printf("only %d values!\n", ii);
    return -2;
    }

  }  /* for it */

  printf("n=%d\n", ii);
  for (it=1; it<=nt; it++) {
    na = a0sum[it];
    a1 = a1sum[it];
    a2 = a2sum[it];
    ma = a1/na;
    sa = sqrt((a2 - a1*a1/na)/(na-1));
    da = sa/sqrt(na-1);
    a2 = sqrt(ma+2*da);
    a1 = (ma > 2*da) ? sqrt(ma-2*da) : 0;
    q = sqrt(sa/ma);
    ma = sqrt(ma);
    printf("%4d  %7.1f  %5.3f  %7.1f  %7.1f\n",
      it, 100*ma, q, 100*a1, 100*a2);
    }
  return 0;
  }

int Ueber( int m, float d )
  {
  int n, i, j, k, l;
  float a, cc[100], zm, zq, csum;
  char *pc;
  if (m > 100)  m = 100;
  if (d <= 0)  d = 10;
  l = strlen(Label);
  if (l < 1) {
    strcpy(Label, "0001:");
    l = 5;
    }
  for (j=0; j<100; j++) cc[j] = 0;
  i = -1;
  n = 0;
  while (fgets(Buf, 4000, Log)) {
    n++;
    if (i < 0) {
      if (!strncmp(Buf,Label,l))  i = 0;
          continue;
      }
    if (strlen(Buf) < 10)  break;
    i++;
    csum = 0;
    pc = strtok(Buf, " \t\r\n");
    for (j=0; j<m; j++) {
      if (!pc) {
        printf("value #%d in line %d not found!\n", j+1, n);
        return -3;
        }
      k = sscanf(pc, "%f", &a);
      if (k != 1)  break;
      cc[j] = a;
      csum += a;
      pc = strtok(NULL, " \t\r\n");
      }
    if (j < m) {
      printf("only %d values in line %d!\n", j, n);
      return -4;
      }
    if (csum <= 0)  continue;
    zm = 0;
    for (j=0; j<m; j++)  zm += cc[j]*(j+0.5)*d;
    zm /= csum;
    zq = 0;
    for (j=0; j<m; j++)  zq += cc[j]*((j+0.5)*d-zm)*((j+0.5)*d-zm);
    zq = sqrt(zq/csum);
    printf("%3d  %6.1f  %6.1f  %6.1f\n", i, zm, zm-zq, zm+zq);
    }
  return 0;
  }

int main( int argc, char *argv[] )
  {
  char fn[256];
  int r, nn;
  float d;
  if (argc < 4) {
    printf("usage: analyse <path> <mode> <count>\n");
    exit(0);
    }
  strcpy(Path, argv[1]);
  strcpy(Mode, argv[2]);
  nn = 0;
  d = 0;
  sscanf(argv[3], "%d,%f", &nn, &d);
  if (argc > 4)  strcpy(Label, argv[4]);
  sprintf(fn, "%s%c%s", Path, '\\', "prtlog.txt");
  Log = fopen(fn, "rb");
  if (!Log) {
    printf("can't read file %s!\n", fn);
    exit(1);
    }
  if      (!strcmp(Mode, "hom-a"))   r = HomA(nn);
  else if (!strcmp(Mode, "taylor"))  r = Taylor(nn, d);
  else if (!strcmp(Mode, "histo"))   r = Histo(nn, d);
  else if (!strcmp(Mode, "ueber"))   r = Ueber(nn, d);
  else {
    printf("unknown mode: %s!\n", Mode);
    exit(0);
    }
  fclose(Log);
  if (r) {
    printf("error during analysis!\n");
    exit(2);
    }
  return 0;
  }

